let option_arg f = function
  None -> f ()
| Some x -> x

let save_to_file s = function
  None -> ()
| Some filename ->
    let filedescr = Unix.openfile filename [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o640 in
    ignore (Unix.write filedescr s 0 (String.length s))

let _ =
  let input_file = ref None in
  let output_yacc = ref None in
  let output_types = ref None in
  let print_yacc = ref false in
  let print_tree = ref false in
  (* Lecture de la ligne de commande *)
  let arg_spec = (Arg.align
    [ "--yacc", Arg.String (fun s -> output_yacc := Some s), "<file> Yacc destination file";
      "--tree", Arg.String (fun s -> output_types := Some s), "<file> Abstract syntax tree module file";
      "--printyacc", Arg.Set print_yacc, " Print compiled yacc on standard output";
      "--printtree", Arg.Set print_tree, " Print compiled abstract syntax tree on standard output";
      "-y", Arg.String (fun s -> output_yacc := Some s), "<file> See --yacc";
      "-t", Arg.String (fun s -> output_types := Some s), "<file> See --tree";
      "-py", Arg.Set print_yacc, " See --printyacc";
      "-pt", Arg.Set print_tree, " See --printtree"
    ]) in
  let arg_usage = "Usage: dccc <file>\nExample: dccc parser.dg -y parser.mly -t parsetree.ml" in
  Arg.parse arg_spec
    (fun s -> if !input_file = None then input_file := Some s else raise (Arg.Bad "Only one input file is allowed"))
    arg_usage;
  (* Ouvrir le fichier d'entrée s'il existe *)
  let filename = option_arg (fun () -> Arg.usage arg_spec arg_usage; exit 1) !input_file in
  if not (Sys.file_exists filename) then print_string ("File not found: "^filename^"\n") else
  let filedescr = Unix.openfile filename [Unix.O_RDONLY] 0 in
  let filechan = Unix.in_channel_of_descr filedescr in
  (* Initialiser le parser *)
  let lexbuf = Lexing.from_channel filechan in
  (* Fonctions de gestion des erreurs *)
  let error s =
    print_string (s^"\n");
    exit 1 in
  let parse_error s = 
    print_string ("File \""^filename^"\", line "^string_of_int lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum^
      ", characters "^string_of_int lexbuf.Lexing.lex_curr_p.Lexing.pos_bol^
      "-"^string_of_int (lexbuf.Lexing.lex_curr_p.Lexing.pos_bol + 1)^":\n"^s^"\n");
    exit 1 in
  (* Parser le fichier *)
  try
    let parsetree = Parser.grammar Lexer.token lexbuf in
    let (_, _, tt, _) as cg = Yacc_compile.compile_grammar parsetree "Ast" in
    let yc = Yacc.yacc_grammar_to_string cg "Ast" in
    let tc = Typetree.compiled_types_to_string (Tree_compile.compile_grammar tt parsetree) in
    if !print_yacc then print_string yc;
    if !print_tree then print_string tc;
    save_to_file yc !output_yacc;
    save_to_file tc !output_types
  with
    Failure s -> parse_error ("Parse error: "^s) (* erreur du lexer *)
  | Parsing.Parse_error -> parse_error "Syntax error." (* erreur du parser *)
  | Parsetree.Parse_error_string s -> parse_error s
  | Tree_compile.TypeUnionError(t, u) ->
      error ("Cannot unify:\n"^Typetree.compiled_type_to_string t^"\nwith:\n"^Typetree.compiled_type_to_string u)
