open Parsetree

type yacc_grammar = yacc_rule list * yacc_starts * yacc_terminals * yacc_precedences

and yacc_starts = (string * string) list

and yacc_terminals = (string * string option) list

and yacc_rule = string * yacc_branch list

and yacc_branch = string list * string

and yacc_precedences = (string * prec) list

(**********************************************************)

module OrderedString = struct
  type t = string
  let compare = compare
end

module StrSet = Set.Make(OrderedString)

module OrderedTT = struct
  type t = string * string option
  let compare = compare
end

module TTSet = Set.Make(OrderedTT)

(**********************************************************)

let rec add_list_to_strset set = function
  [] -> set
| a::tl -> add_list_to_strset (StrSet.add a set) tl

let rec get_branches_tokens aux = function
  [] -> aux
| (toks, _)::tl -> get_branches_tokens (add_list_to_strset aux toks) tl

let rec get_grammar_tokens aux = function
  [] -> aux
| (_, branches)::tl ->
    let nt = get_branches_tokens aux branches in
    get_grammar_tokens nt tl

let rec get_grammar_nt aux = function
  [] -> aux
| (name, _)::tl -> get_grammar_nt (StrSet.add name aux) tl

let get_grammar_t_nt g =
  let toks = get_grammar_tokens StrSet.empty g in
  let nt = get_grammar_nt StrSet.empty g in
  StrSet.diff toks nt, nt

(**********************************************************)

let rec yacc_tokens_to_buf buf = function
  [] -> ()
| t::tl -> Buffer.add_char buf ' '; Buffer.add_string buf t; yacc_tokens_to_buf buf tl

let yacc_branch_to_buf buf vert = function (tokens, semantics) ->
  Buffer.add_char buf '\n';
  if vert then Buffer.add_char buf '|' else Buffer.add_char buf ' ';
  yacc_tokens_to_buf buf tokens;
  Buffer.add_string buf (" { "^semantics^" }")

let rec yacc_branches_to_buf buf vert = function
  [] -> ()
| b::tl -> yacc_branch_to_buf buf vert b; yacc_branches_to_buf buf true tl

let yacc_rule_to_buf buf = function (name, branches) ->
  Buffer.add_string buf (name^":");
  yacc_branches_to_buf buf false branches;
  Buffer.add_string buf ";\n\n"

let yacc_terminal_to_buf buf = function
  name, None -> Buffer.add_string buf ("%token "^name^"\n")
| name, Some t -> Buffer.add_string buf ("%token <"^t^"> "^name^"\n")

let yacc_terminals_to_buf buf terminals =
  TTSet.iter (yacc_terminal_to_buf buf) terminals

let yacc_starts_to_buf buf starts =
  Buffer.add_string buf "%start";
  List.iter (fun (t, ty) -> Buffer.add_char buf ' '; Buffer.add_string buf t) starts;
  Buffer.add_char buf '\n';
  List.iter (fun (t, ty) -> Buffer.add_string buf ("%type <"^ty^"> "^t^"\n")) starts

let prec_to_buf buf = function
  tok, PLeft -> Buffer.add_string buf "%left "; Buffer.add_string buf tok; Buffer.add_char buf '\n'
| tok, PRight -> Buffer.add_string buf "%right "; Buffer.add_string buf tok; Buffer.add_char buf '\n'
| tok, PNonAssoc -> Buffer.add_string buf "%nonassoc "; Buffer.add_string buf tok; Buffer.add_char buf '\n'

let yacc_precedences_to_buf buf precedences =
  List.iter (prec_to_buf buf) precedences

let rec yacc_grammar_to_buf buf = function
  [] -> Buffer.add_string buf "%%\n"
| r::tl -> yacc_rule_to_buf buf r; yacc_grammar_to_buf buf tl

let yacc_grammar_to_buf buf (g, starts, terminals, precedences, treemodule) =
  Buffer.add_string buf ("%{\n  open "^treemodule^"\n  let loc () = symbol_start_pos (), symbol_end_pos ()\n  let locate e = { node = e; loc = loc () }\n%}\n\n");
  yacc_terminals_to_buf buf terminals;
  Buffer.add_char buf '\n';
  yacc_precedences_to_buf buf precedences;
  Buffer.add_char buf '\n';
  yacc_starts_to_buf buf starts;
  Buffer.add_char buf '\n';
  Buffer.add_string buf "%%\n\n";
  yacc_grammar_to_buf buf g

(**********************************************************)

let buf_to_string_f f x =
  let buf = Buffer.create 16 in
  f buf x;
  Buffer.contents buf

let yacc_grammar_to_string (a, b, c, d) tm = buf_to_string_f yacc_grammar_to_buf (a, b, c, d, tm)
let yacc_rule_to_string = buf_to_string_f yacc_rule_to_buf
let yacc_branch_to_string = buf_to_string_f yacc_branch_to_buf
