open Parsetree
open Yacc

type token_semantics =
  SemDirect of int
| SemLabeled of string * token_semantics list
| SemConse of int * int
| SemEmptyList
| SemLocate of token_semantics list
| SemNode of token_semantics

type rule_to_do =
  RBranch of string * tok * tok
| RStar of string * tok
| RSingle of string * tok
| RPlus of string * tok
| ROption of string * tok
| RBang of string * tok * tok

let fresh = let c = ref 0 in fun n ->
  c := !c + 1;
  if String.length n >= 5 && String.sub n 0 5 = "dccc_" then
    String.sub n 0 (String.rindex n '_' + 1)^string_of_int !c
  else
    "dccc_"^n^"_"^(string_of_int !c)

let rec compile_token (tt: Yacc.TTSet.t) name n loc =
  let locate = if loc then fun x -> [SemLocate x] else fun x -> x in function
  TIdent id ->
    let sem = if TTSet.exists (fun (n, t) -> n = id && t = None) tt then [] else
              (*if TTSet.exists (fun (n, t) -> n = id && t <> None) tt then*) [SemDirect n] (*else
              [SemNode (SemDirect n)]*) in
    [id], locate sem, n, []
| TIgnore t ->
    let yt, sem, last, todo = compile_token tt name n false t in
    yt, locate [], last, todo
| TSeq(t, u) ->
    let yt1, sem1, last1, todo1 = compile_token tt name n false t in
    let yt2, sem2, last2, todo2 = compile_token tt name (last1 + 1) false u in
    yt1@yt2, locate (sem1@sem2), last2, todo1@todo2
| TBranch(u, v) ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [RBranch(freshname, u, v)]
| TLabeled(l, t) ->
    let yt, sem, last, todo = compile_token tt name n false t in
    yt, locate [SemLabeled(l, sem)], last, todo
| TStar t ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [RStar(freshname, t)]
| TPlus t ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [RPlus(freshname, t)]
| TOption t ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [ROption(freshname, t)]
| TBang(t, u) ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [RBang(freshname, t, u)]
| TSub t ->
    let freshname = fresh name in
    [freshname], locate [SemDirect n], n, [RSingle(freshname, t)]
| TPrec id ->
    ["%prec "^id], locate [], n, []

let rec string_of_token_semantics (tt: Yacc.TTSet.t) = function
  SemDirect i -> "$"^string_of_int i
| SemLabeled(l, sem) -> let cs = compile_semantics tt sem in if cs = "" then "`"^l else "`"^l^"("^cs^")"
| SemConse(i, j) -> "$"^string_of_int i^"::$"^string_of_int j
| SemEmptyList -> "[]"
| SemLocate sem -> let cs = compile_semantics tt sem in "locate ("^cs^")"
| SemNode sem -> "("^string_of_token_semantics tt sem^").node"

and compile_semantics (tt: Yacc.TTSet.t) = function
  [] -> ""
| s::tl -> let s = string_of_token_semantics tt s in if tl = [] then s else s^", "^compile_semantics tt tl

let rec compile_rule_to_do (tt: Yacc.TTSet.t) = function
  RBranch(name, t, u) ->
    let yt1, sem1, _, todo1 = compile_token tt name 1 false t in
    let yt2, sem2, _, todo2 = compile_token tt name 1 false u in
    (name, [yt1, compile_semantics tt sem1; yt2, compile_semantics tt sem2]), todo1@todo2
| RStar(name, t) ->
    let freshname = fresh name in
    (name, [[freshname; name], "$1::$2"; [], "[]"]), [RSingle(freshname, t)]
| RSingle(name, t) ->
    let yt, sem, _, todo = compile_token tt name 1 false t in
    (name, [yt, compile_semantics tt sem]), todo
| RPlus(name, t) ->
    let freshname = fresh name in
    (name, [[freshname; name], "$1::$2"; [freshname], "[$1]"]), [RSingle(freshname, t)]
| ROption(name, t) ->
    let freshname = fresh name in
    (name, [[freshname], "Some $1"; [], "None"]), [RSingle(freshname, t)]
| RBang(name, t, u) ->
    let freshnamet = fresh name in
    let freshnameu = fresh name in
    (name, [[freshnamet; freshnameu; name], "$1::$3"; [freshnamet], "[$1]"]), [RSingle(freshnamet, t); RSingle(freshnameu, u)]

let rec compile_rules_to_do (tt: Yacc.TTSet.t) = function
  [] -> []
| todo::tl ->
    let ruledone, todonext = compile_rule_to_do tt todo in
    ruledone::(compile_rules_to_do tt todonext)@(compile_rules_to_do tt tl)

let compile_rule (tt: Yacc.TTSet.t) = function name, token ->
  let yt, sem, _, todo = compile_token tt name 1 true token in
  (name, [yt, compile_semantics tt sem])::(compile_rules_to_do tt todo)

let rec get_grammar_starts aux treemodule = function
  [] -> aux
| (GRule(name, _))::tl -> get_grammar_starts ((name, treemodule^"."^name)::aux) treemodule tl
| _::tl -> get_grammar_starts aux treemodule tl

(*********************************************************)

let rec get_token_tokens aux = function
  TIdent s -> TTSet.add (s, None) aux
| TSeq(t, u)
| TBranch(t, u)
| TBang(t, u) -> get_token_tokens (get_token_tokens aux t) u
| TIgnore t
| TStar t
| TPlus t
| TOption t
| TLabeled(_, t)
| TSub t -> get_token_tokens aux t
| TPrec _ -> aux

let rec get_grammar_tokens aux = function
  [] -> aux
| (GRule(_, tok))::tl -> get_grammar_tokens (get_token_tokens aux tok) tl
| _::tl -> get_grammar_tokens aux tl

let rec get_grammar_non_terminals aux = function
  [] -> aux
| (GRule(name, _))::tl -> get_grammar_non_terminals (TTSet.add (name, None) aux) tl
| _::tl -> get_grammar_non_terminals aux tl

let get_grammar_terminals g =
  let toks = get_grammar_tokens TTSet.empty g in
  let nt = get_grammar_non_terminals TTSet.empty g in
  TTSet.diff toks nt

let rec get_grammar_typed_terms aux = function
  [] -> aux
| (GToken(name, tt))::tl ->
    let aux2 = TTSet.filter (fun (n, _) -> n <> name) aux in
    get_grammar_typed_terms (TTSet.add (name, Some tt) aux2) tl
| _::tl -> get_grammar_typed_terms aux tl

(*********************************************************)

let rec get_grammar_precedences = function
  [] -> []
| (GPrec l)::tl -> l@(get_grammar_precedences tl)
| _::tl -> get_grammar_precedences tl

(*********************************************************)

let rec compile_grammar_rules tt = function
  [] -> []
| (GRule(name, tok))::tl -> (compile_rule tt (name, tok))@(compile_grammar_rules tt tl)
| _::tl -> compile_grammar_rules tt tl

let compile_grammar g treemodule =
  let tt = get_grammar_typed_terms (get_grammar_terminals g) g in
  compile_grammar_rules tt g, get_grammar_starts [] treemodule g, tt, get_grammar_precedences g
